extern crate piston_window;
extern crate image as im;
use std::collections::HashMap;
use piston_window::*;
use nalgebra::*;
use rand::Rng;

#[derive(Debug, Clone)]
struct World {
    cells: HashMap<Vector2<i32>, Cell>,
    dimensions: Vector2<i32>
}

impl World {
    fn init(dim: Vector2<i32>) -> World {
        // dbg!("init world", dim);
        World {
            cells: HashMap::new(),
            dimensions: dim
        }
    }

}

#[derive(Debug, Clone, Default)]
struct Cell {
    birth: i32,
    maxage: i32,
    dead: bool,
}

impl Cell {
    fn decomposed(&self, step: i32) -> bool {
        step - self.birth > 100 
    }

    fn dead(&self, step: i32) -> bool {
        step - self.birth > 50 
    }
    
    fn age(&self, step: i32) -> i32 {
        step - self.birth
    }



}



fn main() {
    let (width, height) = (400, 400);


    let mut world = World::init(Vector2::new(width, height));


    let mut rng = rand::thread_rng();

    
    world.cells.insert(Vector2::new(rng.gen_range(0, width), rng.gen_range(0, height)), Cell{birth: 0, maxage: 110, ..Default::default()});
    world.cells.insert(Vector2::new(rng.gen_range(0, width), rng.gen_range(0, height)), Cell{birth: 0, maxage: 110, ..Default::default()});

    let opengl = OpenGL::V3_2;
    let mut window: PistonWindow =
        WindowSettings::new("piston: paint", (width as u32, height as u32))
        .exit_on_esc(true)
        .graphics_api(opengl)
        .build()
        .unwrap();

    let mut canvas = im::ImageBuffer::new(width as u32, height as u32);
    let mut texture_context = TextureContext {
        factory: window.factory.clone(),
        encoder: window.factory.create_command_buffer().into()
    };
    let mut texture: G2dTexture = Texture::from_image(
            &mut texture_context,
            &canvas,
            &TextureSettings::new()
        ).unwrap();

    let mut step = 0;


    let min = Vector2::new(0, 0);
    let max = world.dimensions - Vector2::new(1, 1);

    while let Some(e) = window.next() {


        for (pos, cell) in world.clone().cells {

            // set dead
            if cell.dead(step) {
                let dead_cell = world.cells.get_mut(&pos).unwrap();
                dead_cell.dead = true;
            }

            // decomposed
            if cell.decomposed(step) {
                world.cells.remove(&pos);
                canvas.put_pixel(pos.x as u32, pos.y as u32, im::Rgba([255, 255, 255, 255]));
                continue;

            } 


            if cell.dead {
                canvas.put_pixel(pos.x as u32, pos.y as u32, im::Rgba([255, 255, 0, 255]));
                continue;
            }

            let neighbors: Vec<Vector2<i32>> = vec![
                    Vector2::new(0, 1),
                    Vector2::new(0, -1),
                    Vector2::new(1, 0),
                    Vector2::new(-1, 0),
                    Vector2::new(-1, 1),
                    Vector2::new(-1, -1),
                    Vector2::new(1, 1),
                    Vector2::new(1, -1),
                ]
                .iter()
                .map(move |a| pos+a)
                .map(|x| clamp(x, min, max))
                .filter(|x| rng.gen_bool(0.1))
                .collect();

            canvas.put_pixel(pos.x as u32, pos.y as u32, im::Rgba([0, 0, 0, 255]));

            for n in neighbors {
                world.cells.entry(n).or_insert(Cell{birth: step, maxage: 110, ..Default::default()});
            }

            
        }


        step += 1;
        texture.update(&mut texture_context, &canvas).unwrap();
        window.draw_2d(&e, |c, g, device| {
            texture_context.encoder.flush(device);
            clear([1.0; 4], g);
            image(&texture, c.transform, g);
        });
    }
}